/* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================*/
#include <ncbi.h>
#include <stdarg.h>
#include <blast/blast.h>

static BLAST_Config	blast_config0 = {
		NULL ,
		BLAST_ERR_NONE ,
		(Nlm_VoidPtr (LIBCALLBACK *)()) Nlm_Malloc ,
		NULL ,
		(Nlm_VoidPtr (LIBCALLBACK *)()) Nlm_Realloc ,
		(void (LIBCALLBACK *)()) Nlm_Free ,
		BLAST_REFINED_STATS_DEFAULT ,
		BLAST_CONSISTENCY_DEFAULT ,
		BlastSumPStd ,
		BLAST_SUMP_EPSILON_DEFAULT ,
		BLAST_KARLIN_K_ITER_DEFAULT ,
		BLAST_KARLIN_K_SUMLIMIT_DEFAULT ,
		BLAST_KARLIN_LAMBDA_ITER_DEFAULT ,
		BLAST_KARLIN_LAMBDA_ACCURACY_DEFAULT ,
		BLAST_KARLIN_LAMBDA0_DEFAULT ,
		BLAST_POISSONP_ACCURACY_DEFAULT ,
		BLAST_CONSISTP_ACCURACY_DEFAULT ,
		BLAST_GAPDECAYRATE_DEFAULT ,
		BLAST_WF_EXACT_WORDS_DEFAULT ,
		BLAST_WF_MAP_WORDS_DEFAULT ,
		BLAST_WF_NH_WORDS_DEFAULT ,
		NULL
	};

BLAST_Error LIBCALL
_blast_config_init()
{
	_blast_util_init(&blast_config0);
	blast_config = (*blast_config0.calloc)(sizeof(blast_config0));
	if (blast_config == NULL)
		return blast_errno = BLAST_ERR_MEM;
	*blast_config = blast_config0;
	return BLAST_ERR_NONE;
}

BLAST_ConfigPtr
BlastConfigNew()
{
	return (BLAST_ConfigPtr)BlastCalloc(sizeof(BLAST_Config));
}

void LIBCALL
BlastConfigDestruct(stp)
	BLAST_ConfigPtr	stp;
{
	BlastFree(stp);
}

BLAST_Error LIBCALL
BlastConfigSave (stp)
	BLAST_ConfigPtr	stp;
{
	/* If stp is NULL, make the current state the default */
	if (stp == NULL)
		stp = &blast_config0;
	*stp = *blast_config;
	stp->next = NULL;
	stp->blast_errno = blast_errno;
	return BLAST_ERR_NONE;
}

BLAST_Error LIBCALL
BlastConfigRestore(BLAST_ConfigPtr stp, Nlm_Boolean resterr)
{
	BLAST_ConfigPtr	savenext;

	if (stp == NULL)
		stp = &blast_config0;

	savenext = blast_config->next;
	*blast_config = *stp;
	blast_config->next = savenext;
	if (resterr)
		blast_errno = stp->blast_errno;
	return BLAST_ERR_NONE;
}

BLAST_Error
BlastConfigPush()
{
	BLAST_ConfigPtr	newcp;

	newcp = BlastConfigNew();
	if (newcp == NULL)
		return blast_errno;
	BlastConfigSave(newcp);
	newcp->next = blast_config;
	blast_config = newcp;
	return BLAST_ERR_NONE;
}

BLAST_Error
BlastConfigPop(Nlm_Boolean poperr)
{
	BLAST_ConfigPtr	old;

	if (blast_config->next == NULL)
		return BLAST_ERR_POP;
	old = blast_config;
	blast_config = blast_config->next;
	BlastConfigDestruct(old);
	if (poperr)
		blast_errno = blast_config->blast_errno;
	return BLAST_ERR_NONE;
}
