/* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================*/
#include <ncbi.h>
#include <blast/blast.h>

#define NT_Gap	(BLAST_Letter)0
#define NT_A	(BLAST_Letter)1
#define NT_C	(BLAST_Letter)2
#define NT_M	(BLAST_Letter)3
#define NT_G	(BLAST_Letter)4
#define NT_R	(BLAST_Letter)5
#define NT_S	(BLAST_Letter)6
#define NT_V	(BLAST_Letter)7
#define NT_T	(BLAST_Letter)8
#define NT_W	(BLAST_Letter)9
#define NT_Y	(BLAST_Letter)10
#define NT_H	(BLAST_Letter)11
#define NT_K	(BLAST_Letter)12
#define NT_D	(BLAST_Letter)13
#define NT_B	(BLAST_Letter)14
#define NT_N	(BLAST_Letter)15

static BLAST_Letter nt_Gap[] = { NT_Gap };
static BLAST_Letter nt_A[] = { NT_A };
static BLAST_Letter nt_C[] = { NT_C };
static BLAST_Letter nt_G[] = { NT_G };
static BLAST_Letter nt_T[] = { NT_T };
static BLAST_Letter nt_M[] = { NT_T , NT_C };
static BLAST_Letter nt_R[] = { NT_A , NT_G };
static BLAST_Letter nt_S[] = { NT_C , NT_G };
static BLAST_Letter nt_W[] = { NT_A , NT_T };
static BLAST_Letter nt_Y[] = { NT_C , NT_T };
static BLAST_Letter nt_K[] = { NT_G , NT_T };
static BLAST_Letter nt_V[] = { NT_A , NT_C , NT_G };
static BLAST_Letter nt_H[] = { NT_A , NT_C , NT_T };
static BLAST_Letter nt_D[] = { NT_A , NT_G , NT_T };
static BLAST_Letter nt_B[] = { NT_C , NT_G , NT_T };
static BLAST_Letter nt_N[] = { NT_A , NT_C , NT_G , NT_T };

static BLAST_DegenList ntlist[] = {
			{ DIM(nt_Gap), nt_Gap } ,
			{ DIM(nt_A), nt_A } ,
			{ DIM(nt_C), nt_C } ,
			{ DIM(nt_M), nt_M } ,
			{ DIM(nt_G), nt_G } ,
			{ DIM(nt_R), nt_R } ,
			{ DIM(nt_S), nt_S } ,
			{ DIM(nt_V), nt_V } ,
			{ DIM(nt_T), nt_T } ,
			{ DIM(nt_W), nt_W } ,
			{ DIM(nt_Y), nt_Y } ,
			{ DIM(nt_H), nt_H } ,
			{ DIM(nt_K), nt_K } ,
			{ DIM(nt_D), nt_D } ,
			{ DIM(nt_B), nt_B } ,
			{ DIM(nt_N), nt_N }
		};

static BLAST_DegenList aalist[] = {
			{ 0, NULL }
			/**** NOT DONE ****/
		};

static BLAST_DegenMap ntmap = {
			ntlist ,
			NULL ,
			NULL ,
			NULL
		};

static BLAST_DegenMap aamap = {
			aalist ,
			NULL ,
			NULL ,
			NULL ,
		};


BLAST_DegenMapPtr
BlastDegenMapNew(ap)
	BLAST_AlphabetPtr	ap;
{
	BLAST_AlphabetPtr	stdap;
	BLAST_DegenMapPtr	dmp;
	BLAST_DegenListPtr	dlp;
	BLAST_AlphaMapPtr	amp12, amp21;
	size_t	range;

	if (ap == NULL) {
		blast_errno = BLAST_ERR_INVAL;
		return NULL;
	}

	amp12 = amp21 = NULL;
	if (ap->alphatype == BLAST_ALPHATYPE_AMINO_ACID) {
		if ((stdap = BlastAlphabetFindByID(BLAST_ALPHA_NCBIstdaa)) == NULL)
			return NULL;
		if (ap != stdap && (amp12 = BlastAlphaMapFindCreate(ap, stdap)) == NULL)
			return NULL;
		if (ap != stdap && (amp21 = BlastAlphaMapFindCreate(stdap, ap)) == NULL)
			return NULL;
	}
	else
	if (ap->alphatype == BLAST_ALPHATYPE_NUCLEIC_ACID) {
		if ((stdap = BlastAlphabetFindByID(BLAST_ALPHA_NCBI4na)) == NULL)
			return NULL;
		if (ap != stdap && (amp12 = BlastAlphaMapFindCreate(ap, stdap)) == NULL)
			return NULL;
		if (ap != stdap && (amp21 = BlastAlphaMapFindCreate(stdap, ap)) == NULL)
			return NULL;
	}
	else {
		blast_errno = BLAST_ERR_INVAL;
		return NULL;
	}

	dmp = (BLAST_DegenMapPtr) BlastCalloc(sizeof(*dmp));
	if (dmp == NULL)
		return NULL;

	dmp->ap = ap;

	dmp->amp12 = amp12;
	dmp->amp21 = amp21;

	range = ap->letter_max - ap->letter_min + 1;
	dlp = dmp->_degen = (BLAST_DegenListPtr) BlastCalloc(range * sizeof(*dlp));
	if (dlp == NULL)
		goto Error;

	dmp->degen = dlp - ap->letter_min;
	return dmp;

Error:
	if (dmp != NULL && dmp->_degen != NULL)
		BlastFree(dmp->_degen);
	if (dmp != NULL)
		BlastFree(dmp);
	return NULL;
}


BLAST_Error
BlastDegenMapDestruct(dmp)
	BLAST_DegenMapPtr	dmp;
{
	BLAST_AlphabetPtr	ap;
	BLAST_Letter	ch;
	int	i;

	if (dmp == NULL)
		return blast_errno = BLAST_ERR_INVAL;
	if (dmp == &ntmap || dmp == &aamap)
		return BLAST_ERR_NONE;

	ap = dmp->ap;
	for (i = 0; i < ap->alphasize; ++i) {
		ch = ap->alist[i];
		if (dmp->degen[ch].cnt > 0 && dmp->degen[ch].list != NULL)
			BlastFree(dmp->degen[ch].list);
	}

	if (dmp->_degen != NULL)
		BlastFree(dmp->_degen);

	Nlm_MemSet((Nlm_VoidPtr)dmp, 0, sizeof(BLAST_DegenMap));
	BlastFree(dmp);
	return BLAST_ERR_NONE;
}

/*
	BlastDegenMapFind(ap)

	Create a DegenMap for the specified alphabet, using ntlist or aalist
	as prototypes.
*/
BLAST_DegenMapPtr LIBCALL
BlastDegenMapFind(ap)
	BLAST_AlphabetPtr	ap;
{
	BLAST_DegenMapPtr	dmp;
	BLAST_AlphaMapPtr	amp12, amp21;
	BLAST_DegenListPtr	dlp, dlp2;
	BLAST_DegenListPtr	dglist;
	BLAST_Letter	fch, fch2, tch, tolist[BLAST_ALPHASIZE_MAX];
	size_t	i, j, k, nto;

	if (ap == NULL) {
		blast_errno = BLAST_ERR_INVAL;
		return NULL;
	}
	if (ap->initialized == FALSE) {
		blast_errno = BLAST_ERR_INIT;
		return NULL;
	}
	if (ap->alphasize > BLAST_ALPHASIZE_MAX) {
		blast_errno = BLAST_ERR_ALPHASIZE;
		return NULL;
	}

	dmp = BlastDegenMapNew(ap);
	if (dmp == NULL)
		return NULL;
	amp12 = dmp->amp12;
	amp21 = dmp->amp21;
	if (ap->alphatype == BLAST_ALPHATYPE_AMINO_ACID) {
		dglist = aalist;
	}
	else {
		dglist = ntlist;
	}

	for (i = 0; i < ap->alphasize; ++i) {
		fch = fch2 = ap->alist[i];
		if (amp12 != NULL) {
			if (BlastAlphaMapTst(amp12, fch) == FALSE)
				continue;
			fch2 = BlastAlphaMapChr(amp12, fch);
		}
		dlp = &dmp->degen[fch];
		dlp2 = &dglist[fch2];
		for (j = 0, nto = 0; j < dlp2->cnt; ++j) {
			tch = dlp2->list[j];
			if (amp21 != NULL) {
				if (BlastAlphaMapTst(amp21, tch) == FALSE)
					continue;
				tch = BlastAlphaMapChr(amp21, tch);
			}
			/* maintain a list of distinct letters */
			for (k = 0; k < nto; ++k) {
				if (tolist[k] == tch)
					break;
			}
			if (k == nto) {
				tolist[nto++] = tch;
			}
		}
		dlp->cnt = nto;
		if (nto == 0)
			dlp->list = NULL;
		else {
			dlp->list = (BLAST_LetterPtr) MemDup(tolist, nto*sizeof(tolist[0]));
			if (dlp->list == NULL) {
				BlastDegenMapDestruct(dmp);
				return NULL;
			}
		}
	}

	return dmp;
}
