/* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================*/
#include <ncbi.h>
#include <blast/blast.h>

#ifdef FN
#undef FN
#endif
#ifdef BLAST_STATS
#if 0
#define FN(A) A ## Stats
#else
#define FN(A)	A
#endif
#define BlastWordFinderSelect	BlastWordFinderSelectStats
#define search_nt_orig	search_nt_origStats
#define BlastWordFinder_F1L1	BlastWordFinder_F1L1Stats
#define BlastWordFinder_FnL1	BlastWordFinder_FnL1Stats
#define BlastWordExtend_L1	BlastWordExtend_L1Stats
#define BlastWordFinder_FnL2B4	BlastWordFinder_FnL2B4Stats
#define BlastWordExtend_L2B4	BlastWordExtend_L2B4Stats
#define BlastWordFinder_FnL4B2	BlastWordFinder_FnL4B2Stats
#define BlastWordExtend_L4B2	BlastWordExtend_L4B2Stats
#define BlastHSPExtendNoXFltr_L1	BlastHSPExtendNoXFltr_L1Stats
#else
#define FN(A)	A
#endif

static int LIBCALL FN(BlastWordExtend_L1) PROTO((BLAST_WFContextPtr wfcp, size_t q_off, size_t s_off));
static int LIBCALL FN(BlastWordExtend_L2B4) PROTO((BLAST_WFContextPtr wfcp, size_t q_off, size_t s_off));
static int LIBCALL FN(BlastWordExtend_L4B2) PROTO((BLAST_WFContextPtr wfcp, size_t q_off, size_t s_off));


#ifndef BLAST_STATS
BLAST_Error LIBCALL
BlastWordFinderSearch(wfp, subject)
	BLAST_WordFinderPtr	wfp;
	BLAST_StrPtr	subject;
{
	if (BlastWordFinderSubjectSet(wfp, subject) != BLAST_ERR_NONE)
		return blast_errno;

	return (BLAST_Error)wfp->wordfinder(wfp);
}
#endif /* !BLAST_STATS */


/* BlastWordFinderSelect -- select an appropriate WordFinder function */

int (LIBCALLBACK *
FN( BlastWordFinderSelect )(wfp)) ()
	BLAST_WordFinderPtr	wfp;
{
	if (wfp == NULL) {
		blast_errno = BLAST_ERR_INVAL;
		return NULL;
	}
	if (wfp->type == BLAST_WFTYPE_2)
		return FN( search_nt_orig );
	if (wfp->context_max == 1) {
		switch (wfp->lpb) {
		case 1:
			return FN( BlastWordFinder_F1L1 );
		case 2:
			if (wfp->bpl == 4)
				return FN( BlastWordFinder_FnL2B4 );
			return NULL;
		case 4:
			if (wfp->bpl == 2)
				return FN( BlastWordFinder_FnL4B2 );
			return NULL;
		default:
			return NULL;
		}
		/*NOTREACHED*/
	}

	switch (wfp->lpb) {
	case 1:
		return FN( BlastWordFinder_FnL1 );
	case 2:
		if (wfp->bpl == 4)
			return FN( BlastWordFinder_FnL2B4 );
		return NULL;
	case 4:
		if (wfp->bpl == 2)
			return FN( BlastWordFinder_FnL4B2 );
		return NULL;
	default:
		return NULL;
	}
	/*NOTREACHED*/
}

/*
	search_nt_orig -- an adaptation of the original search_nt() function
	of BLASTN
*/
int LIBCALL
FN( search_nt_orig )(wfp)
	BLAST_WordFinderPtr	wfp;
{
	DFA_StatePtr state;
	DFA_PatlistPtr	plp;
	BLAST_LetterPtr	s;
	BLAST_LetterPtr	s0, sf, sf2;
	BLAST_LetterPtr	q0, qf, q;
	BLAST_Letter	p;
	BLAST_WFContextPtr	wfcp;
	unsigned	left, right;
	unsigned	enclet;
	size_t	q_off;
	DFAPtr	dp;

	if (wfp->state != BLAST_WFSTATE_COMPLETE)
		return blast_errno = BLAST_ERR_WFSTATE;
	if (wfp->type != BLAST_WFTYPE_2)
		return blast_errno = BLAST_ERR_WFTYPE;

	BlastWordFinderSearchEnter(wfp);

	dp = wfp->dp;
	if (dfa_check(dp) != dfaErrNone)
		return blast_errno = BLAST_ERR_DFA;
	state = dp->state0;

	s0 = s = wfp->wfcontext[0].s_str;
	sf = wfp->wfcontext[0].sf;
	sf2 = wfp->wfcontext[0].sf2;

	for (;;) {
		/* The machine "accepts" when a word hit is encountered.  */
		do
			if (s >= sf2)
				goto NormalReturn;
		while (!DFA_ISACCEPTING(state = state->next[*s++]));

		state = (DFA_StatePtr)DFA_UNMUNGED(state);
		plp = &((DFA_AcceptPtr)state)->pl;
		state = ((DFA_AcceptPtr)state)->retState;

		/* Extend each hit in the linked list */
		do {
			wfcp = &wfp->wfcontext[plp->patID.u&BLAST_WORDFINDER_CONTEXT_MASK];
			q_off = (plp->patID.u >> BLAST_WORDFINDER_CONTEXT_SHIFT);
			q0 = wfcp->q_str;
			qf = q0 + wfcp->q_len;
			/* extend to the left */
			p = *(s - wfcp->enclen - 1);
			q = q0 + q_off - (enclet = wfcp->enclet);
			if (RES_L4B2_L3(p) != *--q || q < q0)
				left = 0;
			else
				if (RES_L4B2_L2(p) != *--q || q < q0)
					left = 1;
				else
					if (RES_L4B2_L1(p) != *--q || q < q0)
						left = 2;
					else
						if (RES_L4B2_L0(p) != *--q || q < q0)
							left = 3;
						else
							left = 4;

			/* extend to the right */
			p = *s;
			q = q0 + q_off;
			if (s >= sf || RES_L4B2_L0(p) != *q++ || q > qf)
				right = 0;
			else
				if (RES_L4B2_L1(p) != *q++ || q > qf)
					right = 1;
				else
					if (RES_L4B2_L2(p) != *q++ || q > qf)
						right = 2;
					else
						if (RES_L4B2_L3(p) != *q++ || q > qf)
							right = 3;
						else
							right = 4;

			if (left + right >= wfcp->W - enclet) {
#ifdef BLAST_STATS
				++wfcp->stats.word_hits;
#endif
				if (FN(BlastWordExtend_L4B2)(wfcp, q_off, (size_t)(s - s0)) != 0)
					goto ErrorReturn;
			}
		} while ((plp = plp->chain) != NULL);
	}

NormalReturn:
	BlastWordFinderSearchExit(wfp);
	return 0;

ErrorReturn:
	BlastWordFinderSearchExit(wfp);
	return (int)(blast_errno = BLAST_ERR_APPLICATION);
}

/*----------   search a sequence with 1 Context, 1 Letter per byte  ---------*/
int LIBCALL
FN( BlastWordFinder_F1L1 )(wfp)
	BLAST_WordFinderPtr	wfp;
{
	register DFA_StatePtr state;
	register DFA_PatlistPtr	plp;
	register BLAST_LetterPtr	s;
	register BLAST_WFContextPtr	wfcp;
	BLAST_LetterPtr	s0;
	DFAPtr	dp;

	if (wfp->state != BLAST_WFSTATE_COMPLETE)
		return blast_errno = BLAST_ERR_WFSTATE;
	if (wfp->context_max != 1)
		return blast_errno = BLAST_ERR_INVAL;

	BlastWordFinderSearchEnter(wfp);

	dp = wfp->dp;
	if (dfa_check(dp) != dfaErrNone)
		return blast_errno = BLAST_ERR_DFA;
	state = dp->state0;

	wfcp = &wfp->wfcontext[0];
	s0 = s = wfcp->s_str;

	for (;;) {
		/*
			The machine "accepts" when a word hit is encountered
			or on a sentinel byte
		*/
		while (!DFA_ISACCEPTING(state = state->next[*s++]))
			;

		/* Was a sentinel byte encountered (end of sequence)? */
		if ((state = (DFA_StatePtr)DFA_UNMUNGED(state)) == NULL)
			break;
		plp = &((DFA_AcceptPtr)state)->pl;
		state = ((DFA_AcceptPtr)state)->retState;

		/* Extend each hit in the linked list */
		do {
#ifdef BLAST_STATS
			++wfcp->stats.word_hits;
#endif
			if (FN(BlastWordExtend_L1)(wfcp, (size_t)plp->patID.u, (size_t)(s - s0)) != 0)
				goto ErrorReturn;
		} while ((plp = plp->chain) != NULL);
	}

	BlastWordFinderSearchExit(wfp);
	return 0;

ErrorReturn:
	BlastWordFinderSearchExit(wfp);
	return (int)(blast_errno = BLAST_ERR_APPLICATION);
}

/*----------   search a sequence with N Contexts, 1 Letter per byte  ---------*/
int LIBCALL
FN( BlastWordFinder_FnL1 )(wfp)
	register BLAST_WordFinderPtr	wfp;
{
	register DFA_StatePtr state;
	register DFA_PatlistPtr	plp;
	register BLAST_LetterPtr	s;
	register BLAST_WFContextPtr	wfcp;
	BLAST_LetterPtr	s0;
	size_t		q_off;
	DFAPtr	dp;

	if (wfp->state != BLAST_WFSTATE_COMPLETE)
		return blast_errno = BLAST_ERR_WFSTATE;

	BlastWordFinderSearchEnter(wfp);

	dp = wfp->dp;
	if (dfa_check(dp) != dfaErrNone)
		return blast_errno = BLAST_ERR_DFA;
	state = dp->state0;

	wfcp = &wfp->wfcontext[0];
	s0 = s = wfcp->s_str;

	for (;;) {
		/* The machine "accepts" when a word hit is encountered */
		while (!DFA_ISACCEPTING(state = state->next[*s++]))
			;

		/* Was a sentinel byte encountered (end of sequence)? */
		if ((state = (DFA_StatePtr)DFA_UNMUNGED(state)) == NULL) {
			goto NormalReturn;
		}
		plp = &((DFA_AcceptPtr)state)->pl;
		state = ((DFA_AcceptPtr)state)->retState;

		/* Extend each hit in the linked list */
		do {
			wfcp = &wfp->wfcontext[plp->patID.u & BLAST_WORDFINDER_CONTEXT_MASK];
#ifdef BLAST_STATS
			++wfcp->stats.word_hits;
#endif
			q_off = (plp->patID.u >> BLAST_WORDFINDER_CONTEXT_SHIFT);
			if (FN(BlastWordExtend_L1)(wfcp, (size_t)q_off, (size_t)(s - s0)) != 0) {
				goto ErrorReturn;
			}
		} while ((plp = plp->chain) != NULL);
	}
	/*NOTREACHED*/

NormalReturn:
	BlastWordFinderSearchExit(wfp);
	return 0;

ErrorReturn:
	BlastWordFinderSearchExit(wfp);
	return (int)(blast_errno = BLAST_ERR_APPLICATION);
}

/* BlastWordExtend_L1 -- extend a word-sized hit to a longer match */
static int LIBCALL
FN( BlastWordExtend_L1 )(wfcp, q_off, s_off)
	BLAST_WFContextPtr	wfcp;
	size_t	q_off, s_off;
{
	register BLAST_Diag	diag;
	register BLAST_LetterPtr q, s;
	register BLAST_LetterPtr	q_pos, s_pos;
	register BLAST_Score	sum, score;
	register BLAST_ScoreMat	matrix;
	register BLAST_Score	x;
	register BLAST_Score	X;
	register BLAST_LetterPtr q_beg, q_end;
	BLAST_Offset	off;
	BLAST_OffsetPtr	offp;
	BLAST_Score	leftscore, rightscore, leftsum, rightsum;
	BLAST_LetterPtr	leftq, lefts, rightq, rights;
	BLAST_FilterPtr	ffp;
	BLAST_HSP	hsp;

	offp = &wfcp->dd.diag_level[diag = (BLAST_Diag)s_off - (BLAST_Diag)q_off];
	if ((off = *offp) > q_off)
		return 0;
	if (off == 0)
		*wfcp->dd.diag_stack_top++ = diag;

	/* Find where positive scoring starts & ends within the word hit */
	q_beg = q = (q_end = q_pos = wfcp->q_str + q_off) - wfcp->W;
	s = (s_pos = wfcp->s_str + s_off) - wfcp->W;
	matrix = wfcp->matrix;
	score = sum = 0;
	do
		if ((sum += matrix[*q++][*s++]) > score) {
			score = sum;
			q_end = q;
		}
		else
			if (sum <= 0) {
				sum = 0;
				q_beg = q;
			}
	while (q < q_pos);

	if ((x = -score) < (X = wfcp->X))
		x = X;

	rights = s_pos + ((rightq = q_end) - q_pos);
	lefts = s = s_pos - (q_pos - (leftq = q_beg));
	leftsum = rightsum = rightscore = 0;

Extend_Left:
	/* extend to the left */
	q = leftq;
	s = lefts;
	sum = leftsum;
	do
		if ((sum += matrix[*--q][*--s]) > 0) {
			q_beg = q;
			score += sum;
			/* unroll one iteration of the do-while loop */
			if ((sum = matrix[*--q][*--s]) > 0) {
				q_beg = q;
				score += sum;
				sum = 0;
			}
			if ((x = -score) < X)
				x = X;
		}
	while (sum >= x); /* fails on sentinel byte prefix of either sequence */

	/* extend to the right */
	if (score > rightscore && rightsum > X && -rightscore > X) {
		leftq = q; /* save position that examination for left extension ended */
		lefts = s;
		leftsum = sum;
		leftscore = score;

		q = rightq;
		s = rights;
		sum = rightsum;
		do
			if ((sum += matrix[*q++][*s++]) > 0) {
				q_end = q;
				score += sum;
				/* unroll one iteration of the do-while loop */
				if ((sum = matrix[*q++][*s++]) > 0) {
					q_end = q;
					score += sum;
					sum = 0;
				}
				if ((x = -score) < X)
					x = X;
			}
		while (sum >= x); /* fails on sentinel byte suffix of either sequence */
		rightq = q;
		if (score > leftscore && leftsum > X && -leftscore > X) {
			rights = s;
			rightsum = sum;
			rightscore = score;
			goto Extend_Left;
		}
	}


	/* Record how far this diagonal has been traversed */
	*offp = (rightq - wfcp->q_str);

#ifdef BLAST_STATS
	if (wfcp->stats.high_score < score)
		wfcp->stats.high_score = score;
#endif

	if (score < wfcp->cutoff) { /* Score is not reportable */
#ifdef BLAST_STATS
		++wfcp->stats.failed_extensions;
#endif
		return 0;
	}

	if ((ffp = wfcp->ffp) != NULL) {
#ifdef BLAST_STATS
		++wfcp->stats.passed;
#endif
		hsp.hlp = NULL;
		hsp.next = NULL;
		hsp.vsp = NULL;
		hsp.class = 0;
		hsp.wfcp = wfcp;
		hsp.score = score;
		hsp.kbp = wfcp->kbp;
		hsp.sbp = wfcp->sbp;
		hsp.q_seg.sp = wfcp->query;
		hsp.q_seg.frame = wfcp->query->frame;
		hsp.q_seg.offset = q_beg - wfcp->q_str;
		hsp.s_seg.sp = wfcp->subject;
		hsp.s_seg.frame = wfcp->subject->frame;
		hsp.s_seg.offset = hsp.q_seg.offset + diag;
		hsp.len = hsp.q_seg.len = hsp.s_seg.len = q_end - q_beg;
		return ((BLAST_HSPFltrFnPtr)ffp->func)(ffp, wfcp, &hsp);
	}
	return 0;
}

/* Examine the diagonal rigorously, ignoring drop-off score parameter X */
/* (This was not done by the BLASTP described in the 1990 JMB paper) */
int LIBCALL
FN( BlastHSPExtendNoXFltr_L1 )(ffp, wfcp, hp)
	BLAST_FilterPtr	ffp;
	BLAST_WFContextPtr	wfcp;
	BLAST_HSPPtr	hp;
{
	register BLAST_LetterPtr	q, s;
	register BLAST_Score	sum, score;
	register BLAST_ScoreMat	matrix;
	BLAST_LetterPtr q_beg, q_end;
	BLAST_LetterPtr	q0, qf, s0, sf;
	BLAST_Diag	diag;
	BLAST_HSP	hsp;
	int		rc;

#ifdef BLAST_STATS
	--wfcp->stats.passed;
#endif

	/* Mark this diagonal (in advance) as having been fully examined */
	wfcp->dd.diag_level[diag = (BLAST_Diag)hp->s_seg.offset
			- (BLAST_Diag)hp->q_seg.offset] = BLAST_DIAG_MAX;

	if ((ffp = ffp->next) == NULL)
		return 0;

	matrix = wfcp->matrix;

	q0 = q = wfcp->q_str;
	qf = q + wfcp->q_len;
	s0 = s = wfcp->s_str;
	sf = wfcp->sf;

	/* Start at the beginning of the diagonal */
	if (diag > 0)
		s += diag;
	else
		q -= diag;

	hsp.hlp = NULL;
	hsp.next = NULL;
	hsp.vsp = NULL;
	hsp.class = 0;
	hsp.wfcp = wfcp;
	hsp.kbp = hp->kbp;
	hsp.sbp = hp->sbp;
	hsp.q_seg.sp = hp->q_seg.sp;
	hsp.q_seg.frame = hp->q_seg.frame;
	hsp.s_seg.sp = hp->s_seg.sp;
	hsp.s_seg.frame = hp->s_seg.frame;

	q_beg = q;
	score = sum = 0;
	for (;;) {
		if ((sum += matrix[*q++][*s++]) <= 0) {
			if (score > 0) {
				q = q_end;
				s = s0 + (q_end - q0) + diag;
				if (score >= wfcp->cutoff) {
					/* Record the HSP */
#ifdef BLAST_STATS
					if (wfcp->stats.high_score < score)
						wfcp->stats.high_score = score;
					++wfcp->stats.passed;
#endif
					hsp.score = score;
					hsp.len = hsp.q_seg.len = hsp.s_seg.len = q_end - q_beg;
					hsp.q_seg.offset = q_beg - q0;
					hsp.s_seg.offset = hsp.q_seg.offset + diag;
					if ((rc = ((BLAST_HSPFltrFnPtr)ffp->func)(ffp, wfcp, &hsp)) != 0)
						return rc;
				}
			}
			if (sum < BLAST_SCORE_MIN/2 && (q > qf || s > sf) )
				break; /* Hit a sentinel byte */
			q_beg = q_end = q;
			score = sum = 0;
		}
		else
			if (sum > score) {
				score = sum;
				q_end = q;
			}
	} /* for(;;) */

	return 0;
}

/*-- search a sequence, N Contexts, 2 Letters per byte, 4 bits per letter --*/
int LIBCALL
FN( BlastWordFinder_FnL2B4 )(wfp)
	register BLAST_WordFinderPtr	wfp;
{
	register DFA_StatePtr state;
	register BLAST_LetterPtr	s, sf2;
	register DFA_PatlistPtr	plp;
	register BLAST_WFContextPtr	wfcp;
	BLAST_LetterPtr	s0, sf;
	BLAST_WFHashHitPtr	hhp;
	size_t		q_off;
	unsigned	remain;
	DFAPtr	dp;

	if (wfp->state != BLAST_WFSTATE_COMPLETE)
		return blast_errno = BLAST_ERR_WFSTATE;
	if (wfp->lpb != 2 || wfp->ap->bpl != 4)
		return blast_errno = BLAST_ERR_ENCODING;

	BlastWordFinderSearchEnter(wfp);

	dp = wfp->dp;
	if (dfa_check(dp) != dfaErrNone)
		return blast_errno = BLAST_ERR_DFA;
	state = dp->state0;

	wfcp = &wfp->wfcontext[0];
	s = s0 = wfcp->s_str;
	sf = wfcp->sf;
	sf2 = wfcp->sf2;
	remain = 4*wfcp->remain;
	remain = 0xff >> (8-remain);
	remain = remain << (8-remain);

	for (;;) {
		/* The machine "accepts" when a word hit is encountered */
		do
			if (s >= sf2)
				goto NormalReturn;
		while (!DFA_ISACCEPTING(state = state->next[*s++]));

		state = (DFA_StatePtr)DFA_UNMUNGED(state);
		plp = &((DFA_AcceptPtr)state)->pl;
		state = ((DFA_AcceptPtr)state)->retState;

		/* Extend each hit in the linked list */
		do {
			hhp = (BLAST_WFHashHitPtr)&plp->patID;
			if ((*s & hhp->rmask) != hhp->rchr)
				continue;
			if ((s[-(int)hhp->enclen] & hhp->lmask) != hhp->lchr)
				continue;
			if (hhp->lmask && s < s0 + hhp->enclen)
				continue;
			if (hhp->rmask && s >= sf2) {
				if (s >= sf || remain < hhp->rmask)
					continue;
			}
			wfcp = &wfp->wfcontext[hhp->u & BLAST_WORDFINDER_CONTEXT_MASK];
#ifdef BLAST_STATS
			++wfcp->stats.word_hits;
#endif
			q_off = (hhp->u >> BLAST_WORDFINDER_CONTEXT_SHIFT);
			if (FN(BlastWordExtend_L2B4)(wfcp, q_off, (size_t)(s - s0)) != 0)
				goto ErrorReturn;
		} while ((plp = plp->chain) != NULL);
	}
	/*NOTREACHED*/

NormalReturn:
	BlastWordFinderSearchExit(wfp);
	return 0;

ErrorReturn:
	BlastWordFinderSearchExit(wfp);
	return (int)(blast_errno = BLAST_ERR_APPLICATION);
}

static int LIBCALL
FN( BlastWordExtend_L2B4 )(wfcp, q_off, s_off)
	BLAST_WFContextPtr	wfcp;
	size_t	q_off, s_off;
{
	register BLAST_Diag	diag;
	register BLAST_LetterPtr	q;
	register BLAST_ScoreMat	matrix;
	register BLAST_Score	sum, score;
	BLAST_Offset	off;
	BLAST_OffsetPtr	offp;
	BLAST_Letter	ch;
	BLAST_LetterPtr	q0, s0, sf2, q_beg, q_end, s_end, s;
	BLAST_Score	x, X;
	BLAST_HSP	hsp;
	BLAST_FilterPtr	ffp;
	int		remain;

	offp = &wfcp->dd.diag_level[diag = s_off*2 - q_off];
	if ((off = *offp) > q_off)
		return 0;
	if (off == 0)
		*wfcp->dd.diag_stack_top++ = diag;

	matrix = wfcp->matrix;
	q0 = wfcp->q_str;
	s0 = wfcp->s_str;
	/* sf2 points just after the last fully packed byte */
	sf2 = wfcp->sf2;

	q = q_beg = q_end = q0 + q_off;
	s = s_end = s0 + s_off;

	/* Find where positive scoring starts & ends within the word hit */
	score = sum = 0;

	x = X = wfcp->X;

	/* extend to the left */
	do {
		ch = *--s;
		if ((sum += matrix[*--q][RES_L2B4_L1(ch)]) > 0) {
			q_beg = q;
			score += sum;
			sum = 0;
			if ((x = -score) < X)
				x = X;
		}
		else
			if (sum < x)
				break;
		if ((sum += matrix[*--q][RES_L2B4_L0(ch)]) > 0) {
			q_beg = q;
			score += sum;
			sum = 0;
			if ((x = -score) < X)
				x = X;
		}
		else
			if (sum < x)
				break;
	} while (s > s0);

	/* extend to the right */
	q = q_end;
	s = s_end;
	sum = 0;
	while (s < sf2) {
		ch = *s++;
		if ((sum += matrix[*q++][RES_L2B4_L0(ch)]) > 0) {
			q_end = q;
			score += sum;
			sum = 0;
			if ((x = -score) < X)
				x = X;
		}
		else
			if (sum < x)
				break;
		if ((sum += matrix[*q++][RES_L2B4_L1(ch)]) > 0) {
			q_end = q;
			score += sum;
			sum = 0;
			if ((x = -score) < X)
				x = X;
		}
		else
			if (sum < x)
				break;
	}

	/* extend into the final, partially packed byte (if one exists) */
	if (s == sf2 && sum >= x && wfcp->remain > 0) {
		remain = wfcp->remain;
		ch = *sf2;
		do {
			if ((sum += matrix[*q++][RES_L2B4_L0(ch)]) > 0) {
				q_end = q;
				score += sum;
				sum = 0;
				if ((x = -score) < X)
					x = X;
			}
			else
				if (sum < x)
					break;
			ch <<= 4;
		} while (--remain > 0);
	}

	/* Record how far this diagonal has been traversed */
	*offp = (q - q0);

#ifdef BLAST_STATS
	if (wfcp->stats.high_score < score)
		wfcp->stats.high_score = score;
#endif

	if (score < wfcp->cutoff) { /* Score is not reportable */
#ifdef BLAST_STATS
		++wfcp->stats.failed_extensions;
#endif
		return 0;
	}

	if ((ffp = wfcp->ffp) != NULL) {
#ifdef BLAST_STATS
		++wfcp->stats.passed;
#endif
		hsp.hlp = NULL;
		hsp.next = NULL;
		hsp.vsp = NULL;
		hsp.class = 0;
		hsp.wfcp = wfcp;
		hsp.score = score;
		hsp.kbp = wfcp->kbp;
		hsp.sbp = wfcp->sbp;
		hsp.q_seg.sp = wfcp->query;
		hsp.q_seg.frame = wfcp->query->frame;
		hsp.q_seg.offset = q_beg - q0;
		hsp.s_seg.sp = wfcp->subject;
		hsp.s_seg.frame = wfcp->subject->frame;
		hsp.s_seg.offset = s_off*2 - (q_off - (q_beg - q0));
		hsp.len = hsp.q_seg.len = hsp.s_seg.len = q_end - q_beg;
		return ((BLAST_HSPFltrFnPtr)ffp->func)(ffp, wfcp, &hsp);
	}

	return 0;
}

/*-- search a sequence, N Contexts, 4 Letters per byte, 2 bits per letter --*/
int LIBCALL
FN( BlastWordFinder_FnL4B2 )(wfp)
	register BLAST_WordFinderPtr	wfp;
{
	register DFA_StatePtr state;
	register BLAST_LetterPtr	s, sf2;
	register DFA_PatlistPtr	plp;
	register BLAST_WFContextPtr	wfcp;
	BLAST_LetterPtr	s0, sf;
	BLAST_WFHashHitPtr	hhp;
	size_t		q_off;
	unsigned	remain;
	DFAPtr	dp;

	if (wfp->state != BLAST_WFSTATE_COMPLETE)
		return blast_errno = BLAST_ERR_WFSTATE;
	if (wfp->lpb != 4 || wfp->ap->bpl != 2)
		return blast_errno = BLAST_ERR_ENCODING;

	BlastWordFinderSearchEnter(wfp);

	dp = wfp->dp;
	if (dfa_check(dp) != dfaErrNone)
		return blast_errno = BLAST_ERR_DFA;
	state = dp->state0;

	wfcp = &wfp->wfcontext[0];
	s = s0 = wfcp->s_str;
	sf = wfcp->sf;
	/* sf2 points just after the last fully packed letter */
	sf2 = wfcp->sf2;
	remain = 2*wfcp->remain;
	remain = 0xff >> (8-remain);
	remain = remain << (8-remain);

	for (;;) {
		/* The machine "accepts" when a word hit is encountered */
		do
			if (s >= sf2)
				goto NormalReturn;
		while (!DFA_ISACCEPTING(state = state->next[*s++]));

		state = (DFA_StatePtr)DFA_UNMUNGED(state);
		plp = &((DFA_AcceptPtr)state)->pl;
		state = ((DFA_AcceptPtr)state)->retState;

		/* Extend each hit in the linked list */
		do {
			hhp = (BLAST_WFHashHitPtr)&plp->patID;
			if ((*s & hhp->rmask) != hhp->rchr)
				continue;
			if ((s[-(int)hhp->enclen] & hhp->lmask) != hhp->lchr)
				continue;
			if (hhp->lmask && s < s0 + hhp->enclen)
				continue;
			if (hhp->rmask && s >= sf2) {
				if (s >= sf || remain < hhp->rmask)
					continue;
			}
			wfcp = &wfp->wfcontext[hhp->u & BLAST_WORDFINDER_CONTEXT_MASK];
#ifdef BLAST_STATS
			++wfcp->stats.word_hits;
#endif
			q_off = (hhp->u >> BLAST_WORDFINDER_CONTEXT_SHIFT);
			if (FN(BlastWordExtend_L4B2)(wfcp, q_off, (size_t)(s - s0)) != 0)
				goto ErrorReturn;
		} while ((plp = plp->chain) != NULL);
	}
	/*NOTREACHED*/

NormalReturn:
	BlastWordFinderSearchExit(wfp);
	return 0;

ErrorReturn:
	BlastWordFinderSearchExit(wfp);
	return (int)(blast_errno = BLAST_ERR_APPLICATION);
}

/* BlastWordExtend_L4B2 -- extend a word hit */
static int LIBCALL
FN( BlastWordExtend_L4B2 )(wfcp, q_off, s_off)
	BLAST_WFContextPtr	wfcp;
	size_t	q_off, s_off;
{
	register BLAST_LetterPtr	q;
	register BLAST_ScoreMat	matrix;
	register BLAST_Score	sum, score;
	register BLAST_Diag		diag;
	BLAST_Offset	off;
	BLAST_OffsetPtr	offp;
	BLAST_Letter	ch;
	BLAST_LetterPtr	q0, s0, sf2, q_beg, q_end, s_end, s;
	BLAST_Score	x, X;
	BLAST_FilterPtr	ffp;
	BLAST_HSP	hsp;
	int		remain;

	offp = &wfcp->dd.diag_level[diag = s_off*4 - q_off];
	if ((off = *offp) > q_off)
		return 0;
	if (off == 0)
		*wfcp->dd.diag_stack_top++ = diag;

	matrix = wfcp->matrix;
	q0 = wfcp->q_str;
	s0 = wfcp->s_str;
	/* sf2 points just after the last fully packed byte */
	sf2 = wfcp->sf2;

	q = q_beg = q_end = q0 + q_off;
	s = s_end = s0 + s_off;

	/* Find where positive scoring starts & ends within the word hit */
	score = sum = 0;

	x = X = wfcp->X;

	/* extend to the left */
	do {
		ch = *--s;
		if ((sum += matrix[*--q][RES_L4B2_L3(ch)]) > 0) {
			q_beg = q;
			score += sum;
			sum = 0;
			if ((x = -score) < X)
				x = X;
		}
		else
			if (sum < x)
				break;
		if ((sum += matrix[*--q][RES_L4B2_L2(ch)]) > 0) {
			q_beg = q;
			score += sum;
			sum = 0;
			if ((x = -score) < X)
				x = X;
		}
		else
			if (sum < x)
				break;
		if ((sum += matrix[*--q][RES_L4B2_L1(ch)]) > 0) {
			q_beg = q;
			score += sum;
			sum = 0;
			if ((x = -score) < X)
				x = X;
		}
		else
			if (sum < x)
				break;
		if ((sum += matrix[*--q][RES_L4B2_L0(ch)]) > 0) {
			q_beg = q;
			score += sum;
			sum = 0;
			if ((x = -score) < X)
				x = X;
		}
		else
			if (sum < x)
				break;
	} while (s > s0);

	/* extend to the right */
	q = q_end;
	s = s_end;
	sum = 0;
	while (s < sf2) {
		ch = *s++;
		if ((sum += matrix[*q++][RES_L4B2_L0(ch)]) > 0) {
			q_end = q;
			score += sum;
			sum = 0;
			if ((x = -score) < X)
				x = X;
		}
		else
			if (sum < x)
				break;
		if ((sum += matrix[*q++][RES_L4B2_L1(ch)]) > 0) {
			q_end = q;
			score += sum;
			sum = 0;
			if ((x = -score) < X)
				x = X;
		}
		else
			if (sum < x)
				break;
		if ((sum += matrix[*q++][RES_L4B2_L2(ch)]) > 0) {
			q_end = q;
			score += sum;
			sum = 0;
			if ((x = -score) < X)
				x = X;
		}
		else
			if (sum < x)
				break;
		if ((sum += matrix[*q++][RES_L4B2_L3(ch)]) > 0) {
			q_end = q;
			score += sum;
			sum = 0;
			if ((x = -score) < X)
				x = X;
		}
		else
			if (sum < x)
				break;
	}

	/* extend into the final, partially packed byte (if one exists) */
	if (s == sf2 && sum >= x && wfcp->remain > 0) {
		remain = wfcp->remain;
		ch = *sf2;
		do {
			if ((sum += matrix[*q++][RES_L4B2_L0(ch)]) > 0) {
				q_end = q;
				score += sum;
				sum = 0;
				if ((x = -score) < X)
					x = X;
			}
			else
				if (sum < x)
					break;
			ch <<= 2;
		} while (--remain > 0);
	}

	/* Record how far this diagonal has been traversed */
	*offp = (q - wfcp->q_str);

#ifdef BLAST_STATS
	if (wfcp->stats.high_score < score)
		wfcp->stats.high_score = score;
#endif

	if (score < wfcp->cutoff) { /* Score is not reportable */
#ifdef BLAST_STATS
		++wfcp->stats.failed_extensions;
#endif
		return 0;
	}

	if ((ffp = wfcp->ffp) != NULL) {
#ifdef BLAST_STATS
		++wfcp->stats.passed;
#endif
		hsp.hlp = NULL;
		hsp.next = NULL;
		hsp.vsp = NULL;
		hsp.class = 0;
		hsp.wfcp = wfcp;
		hsp.score = score;
		hsp.kbp = wfcp->kbp;
		hsp.sbp = wfcp->sbp;
		hsp.q_seg.sp = wfcp->query;
		hsp.q_seg.frame = wfcp->query->frame;
		hsp.q_seg.offset = q_beg - q0;
		hsp.s_seg.sp = wfcp->subject;
		hsp.s_seg.frame = wfcp->subject->frame;
		hsp.s_seg.offset = s_off*4 - (q_off - (q_beg - q0));
		hsp.len = hsp.q_seg.len = hsp.s_seg.len = q_end - q_beg;
		return ((BLAST_HSPFltrFnPtr)ffp->func)(ffp, wfcp, &hsp);
	}

	return 0;
}
