#include <ncbi.h>
#include <blast/blast.h>

/******************  BLAST_HitList Comparison Routines  *********************/
ValNodePtr LIBCALL
BlastHitListCmpCriterionAdd(vnpp, crit)
	ValNodePtr	PNTR vnpp;
	BLAST_HLCmp	crit;
{
	ValNodePtr	vnp;

	if (vnpp == NULL)
		return NULL;

	for (vnp = *vnpp; vnp != NULL; vnp = vnp->next) {
		if (vnp->choice == crit)
			return vnp;
	}

	switch (crit) {
	case BLAST_HLCMP_PVALUE:
		return ValNodeAddFunction(vnpp, crit, (FnPtr)BlastHitListCmpByPvalue);
	case BLAST_HLCMP_EVALUE:
		return ValNodeAddFunction(vnpp, crit, (FnPtr)BlastHitListCmpByEvalue);
	case BLAST_HLCMP_SCORE:
		return ValNodeAddFunction(vnpp, crit, (FnPtr)BlastHitListCmpByScore);
	case BLAST_HLCMP_COUNT:
		return ValNodeAddFunction(vnpp, crit, (FnPtr)BlastHitListCmpByCount);
	case BLAST_HLCMP_LENGTH:
		return ValNodeAddFunction(vnpp, crit, (FnPtr)BlastHitListCmpByLength);
	case BLAST_HLCMP_HIGHSCORE:
		return ValNodeAddFunction(vnpp, crit, (FnPtr)BlastHitListCmpByHighScore);
	case BLAST_HLCMP_TOTALSCORE:
		return ValNodeAddFunction(vnpp, crit, (FnPtr)BlastHitListCmpByTotalScore);
	default:
		return NULL;
	}
	/*NOTREACHED*/
}

int LIBCALL
BlastHitListCmpStd(hli, hlj, vnp)
	BLAST_HitListPtr	hli, hlj;
	ValNodePtr	vnp;
{
	register int (LIBCALLBACK *func)();
	register BLAST_HSPPtr	hi, hj;
	register int	i;

	if (hli == NULL)
		return (hlj != NULL ? 1 : 0);
	if (hlj == NULL)
		return -1;

	hi = hli->best_hsp;
	hj = hlj->best_hsp;
	if (hi == NULL)
		return (hj != NULL ? 1 : 0);
	if (hj == NULL)
		return -1;

	for (; vnp != NULL; vnp = vnp->next) {
		func = (int (LIBCALLBACK *)())vnp->data.funcvalue;
		i = (*func)(&hli, &hlj);
		if (i != 0)
			return i;
	}

	return 0;
}

int LIBCALL
BlastHitListPtrCmpStd(hli, hlj, vnp)
	BLAST_HitListPtr	PNTR hli, PNTR hlj;
	ValNodePtr	vnp;
{
	register int (LIBCALLBACK *func)();
	register BLAST_HSPPtr	hi, hj;
	register int	i;

	if (hli == NULL)
		return (hlj != NULL ? 1 : 0);
	if (hlj == NULL)
		return -1;

	hi = (*hli)->best_hsp;
	hj = (*hlj)->best_hsp;
	if (hi == NULL)
		return (hj != NULL ? 1 : 0);
	if (hj == NULL)
		return -1;

	for (; vnp != NULL; vnp = vnp->next) {
		func = (int (LIBCALLBACK *)())vnp->data.funcvalue;
		i = (*func)(hli, hlj);
		if (i != 0)
			return i;
	}

	return 0;
}

void LIBCALL
BlastHitListSort(hlpp, vnp)
	BLAST_HitListPtr PNTR	hlpp;
	ValNodePtr	vnp;
{
	LinkSort1((VoidPtr PNTR)hlpp, offsetof(BLAST_HitList,next), BlastHitListPtrCmpStd, vnp);
}

int LIBCALL
BlastHitListCmpByPvalue(hli, hlj)
	BLAST_HitListPtr	PNTR hli, PNTR hlj;
{
	register BLAST_HSPPtr	hi, hj;

	hi = (*hli)->best_hsp;
	hj = (*hlj)->best_hsp;

	if (hi->pvalue < hj->pvalue)
		return -1;
	if (hi->pvalue > hj->pvalue)
		return 1;
	return 0;
}

int LIBCALL
BlastHitListCmpByEvalue(hli, hlj)
	BLAST_HitListPtr	PNTR hli, PNTR hlj;
{
	register BLAST_HSPPtr	hi, hj;

	hi = (*hli)->best_hsp;
	hj = (*hlj)->best_hsp;

	if (hi->evalue < hj->evalue)
		return -1;
	if (hi->evalue > hj->evalue)
		return 1;
	return 0;
}

int LIBCALL
BlastHitListCmpByLength(hli, hlj)
	BLAST_HitListPtr	PNTR hli, PNTR hlj;
{
	BLAST_HSPPtr	hp;
	BLAST_Offset	imax, jmax;

	hp = (*hli)->best_hsp;
	imax = MAX(hp->q_seg.len, hp->s_seg.len);

	hp = (*hlj)->best_hsp;
	jmax = MAX(hp->q_seg.len, hp->s_seg.len);

	if (imax < jmax)
		return -1;
	if (imax > jmax)
		return 1;
	return 0;
}

int LIBCALL
BlastHitListCmpByCount(hli, hlj)
	BLAST_HitListPtr	PNTR hli, PNTR hlj;
{
	register int	i, j;

	i = BlastHSPCount(*(*hli)->hpp);
	j = BlastHSPCount(*(*hlj)->hpp);

	return j - i;
}

int LIBCALL
BlastHitListCmpByHighScore(hli, hlj)
	BLAST_HitListPtr	PNTR hli, PNTR hlj;
{
	register BLAST_Score	si, sj, sdiff;

	si = (*(*hli)->hpp)->score;
	sj = (*(*hlj)->hpp)->score;

	sdiff = sj - si;
	return SIGN(sdiff);
}

int LIBCALL
BlastHitListCmpByScore(hli, hlj)
	BLAST_HitListPtr	PNTR hli, PNTR hlj;
{
	register BLAST_HSPPtr	hpi, hpj;
	BLAST_Score	si, sj, sdiff;

	hpi = *(*hli)->hpp;
	hpj = *(*hlj)->hpp;

	while (hpi != NULL && hpj != NULL) {
		si = hpi->score;
		sj = hpj->score;
		sdiff = sj - si;
		if (sdiff != 0)
			return (sdiff > 0 ? 1 : -1);
		hpi = hpi->next;
		hpj = hpj->next;
	}
	if (hpi == NULL)
		return (hpj != NULL ? 1 : 0);
	return -1;
}

int LIBCALL
BlastHitListCmpByTotalScore(hli, hlj)
	BLAST_HitListPtr	PNTR hli, PNTR hlj;
{
	register BLAST_HSPPtr	hp;
	register BLAST_Score	s;
	BLAST_Score	si, sj, sdiff;

	for (hp = *(*hli)->hpp, s = 0; hp != NULL; hp = hp->next)
		s += hp->score;
	si = s;

	for (hp = *(*hlj)->hpp, s = 0; hp != NULL; hp = hp->next)
		s += hp->score;
	sj = s;

	sdiff = sj - si;
	return SIGN(sdiff);
}

/******************  BLAST_HSP Comparison Routines  *********************/
ValNodePtr LIBCALL
BlastHSPCmpCriterionAdd(vnpp, crit)
	ValNodePtr	PNTR vnpp;
	BLAST_HSPCmp	crit;
{
	ValNodePtr	vnp;

	if (vnpp == NULL)
		return NULL;

	for (vnp = *vnpp; vnp != NULL; vnp = vnp->next) {
		if (vnp->choice == crit)
			return vnp;
	}

	switch (crit) {
	case BLAST_HSPCMP_PVALUE:
		return ValNodeAddFunction(vnpp, crit, (FnPtr)BlastHSPCmpByPvalue);
	case BLAST_HSPCMP_EVALUE:
		return ValNodeAddFunction(vnpp, crit, (FnPtr)BlastHSPCmpByEvalue);
	case BLAST_HSPCMP_FRAME:
		return ValNodeAddFunction(vnpp, crit, (FnPtr)BlastHSPCmpByFrame);
	case BLAST_HSPCMP_SCORE:
		return ValNodeAddFunction(vnpp, crit, (FnPtr)BlastHSPCmpByScore);
	case BLAST_HSPCMP_NORMSCORE:
		return ValNodeAddFunction(vnpp, crit, (FnPtr)BlastHSPCmpByNormScore);
	case BLAST_HSPCMP_N:
		return ValNodeAddFunction(vnpp, crit, (FnPtr)BlastHSPCmpByN);
	case BLAST_HSPCMP_LENGTH:
		return ValNodeAddFunction(vnpp, crit, (FnPtr)BlastHSPCmpByLength);
	case BLAST_HSPCMP_OFFSET:
		return ValNodeAddFunction(vnpp, crit, (FnPtr)BlastHSPCmpByOffset);
	default:
		return NULL;
	}
	/*NOTREACHED*/
}

int LIBCALL
BlastHSPCmpStd(hspi, hspj, vnp)
	BLAST_HSPPtr	hspi, hspj;
	ValNodePtr	vnp;
{
	register int (LIBCALLBACK *func)();
	register int	i;

	if (hspi == NULL)
		return (hspj != NULL ? 1 : 0);
	if (hspj == NULL)
		return -1;

	for (; vnp != NULL; vnp = vnp->next) {
		func = (int (LIBCALLBACK *)())vnp->data.funcvalue;
		i = (*func)(&hspi, &hspj);
		if (i != 0)
			return i;
	}

	return 0;
}

int LIBCALL
BlastHSPPtrCmpStd(hspi, hspj, vnp)
	BLAST_HSPPtr	PNTR hspi, PNTR hspj;
	ValNodePtr	vnp;
{
	register int (LIBCALLBACK *func)();
	register int	i;

	if (hspi == NULL)
		return (hspj != NULL ? 1 : 0);
	if (hspj == NULL)
		return -1;

	for (; vnp != NULL; vnp = vnp->next) {
		func = (int (LIBCALLBACK *)())vnp->data.funcvalue;
		i = (*func)(hspi, hspj);
		if (i != 0)
			return i;
	}

	return 0;
}

void LIBCALL
BlastHSPSort(hspp, vnp)
	BLAST_HSPPtr PNTR	hspp;
	ValNodePtr	vnp;
{
	LinkSort1((VoidPtr PNTR)hspp, offsetof(BLAST_HSP,next), BlastHSPPtrCmpStd, vnp);
}

int LIBCALL
BlastHSPCmpByPvalue(hspi, hspj)
	BLAST_HSPPtr	PNTR hspi, PNTR hspj;
{
	register double	xi, xj, ratio;

	xi = (*hspi)->pvalue;
	xj = (*hspj)->pvalue;

	if (xi != 0. && xj != 0.) {
		ratio = xi / xj;
		if (ratio < 1.00001 && ratio > 0.99999)
			return 0;
	}

	if (xi < xj)
		return -1;
	if (xi > xj)
		return 1;
	return 0;
}

int LIBCALL
BlastHSPCmpByEvalue(hspi, hspj)
	BLAST_HSPPtr	PNTR hspi, PNTR hspj;
{
	register double	xi, xj, ratio;

	xi = (*hspi)->evalue;
	xj = (*hspj)->evalue;

	if (xi > 1. && xj < 1.) /* avoids possible arithmetic overflow */
		return 1;

	if (xi != 0. && xj != 0.) {
		ratio = xi / xj;
		if (ratio < 1.00001 && ratio > 0.99999)
			return 0;
	}

	if (xi < xj)
		return -1;
	if (xi > xj)
		return 1;
	return 0;
}

int LIBCALL
BlastHSPCmpByFrame(hspi, hspj)
	BLAST_HSPPtr	PNTR hspi, PNTR hspj;
{
	int	fi, fj;

	fi = (*hspi)->q_seg.frame;
	fj = (*hspj)->q_seg.frame;
	if (fi != fj) {
		fi = SIGN(fi);
		fj = SIGN(fj);
		if (fi > fj)
			return -1;
		if (fi < fj)
			return 1;
	}

	fi = (*hspi)->s_seg.frame;
	fj = (*hspj)->s_seg.frame;
	if (fi != fj) {
		fi = SIGN(fi);
		fj = SIGN(fj);
		if (fi > fj)
			return -1;
		if (fi < fj)
			return 1;
	}

	return 0;
}

int LIBCALL
BlastHSPCmpByScore(hspi, hspj)
	BLAST_HSPPtr	PNTR hspi, PNTR hspj;
{
	BLAST_Score	si, sj, sdiff;

	si = (*hspi)->score;
	sj = (*hspj)->score;
	sdiff = sj - si;
	return SIGN(sdiff);
}

int LIBCALL
BlastHSPCmpByNormScore(hspi, hspj)
	BLAST_HSPPtr	PNTR hspi, PNTR hspj;
{
	BLAST_Score	si, sj, sdiff;
	BLAST_KarlinBlkPtr	ki, kj;
	register double	xi, xj, ratio;

	si = (*hspi)->score;
	ki = (*hspi)->kbp;
	sj = (*hspj)->score;
	kj = (*hspj)->kbp;
	if (ki == kj) {
		sdiff = sj - si;
		return SIGN(sdiff);
	}

	xi = ki->Lambda * si;
	xj = kj->Lambda * sj;

	if (xi != 0. && xj != 0.) {
		ratio = xi / xj;
		if (ratio < 1.00001 && ratio > 0.99999)
			return 0;
	}

	if (xi > xj)
		return -1;
	if (xi < xj)
		return 1;
	return 0;
}

int LIBCALL
BlastHSPCmpByLength(hspi, hspj)
	BLAST_HSPPtr	PNTR hspi, PNTR hspj;
{
	BLAST_HSPPtr	hp;
	BLAST_Offset	imax, jmax;

	hp = *hspi;
	imax = MAX(hp->q_seg.len, hp->s_seg.len);
	hp = *hspj;
	jmax = MAX(hp->q_seg.len, hp->s_seg.len);

	if (imax < jmax)
		return -1;
	if (imax > jmax)
		return 1;
	return 0;
}

int LIBCALL
BlastHSPCmpByOffset(hspi, hspj)
	BLAST_HSPPtr	PNTR hspi, PNTR hspj;
{
	register unsigned long	oi, oj;

	oi = (*hspi)->q_seg.offset;
	oj = (*hspj)->q_seg.offset;
	if (oi > oj)
		return 1;
	if (oi < oj)
		return -1;

	oi = (*hspi)->s_seg.offset;
	oj = (*hspj)->s_seg.offset;
	if (oi > oj)
		return 1;
	if (oi < oj)
		return -1;

	return 0;
}

int LIBCALL
BlastHSPCmpByN(hspi, hspj)
	BLAST_HSPPtr	PNTR hspi, PNTR hspj;
{
	register int	i, j;

	i = (*hspi)->n;
	j = (*hspj)->n;

	return j - i;
}
